#pragma once

#include <time.h>
#include <string>

#include "../graph.cpp"
#include "../digraph.cpp"
#include "../graphio.cpp"
#include "beamSearch.cpp"
#include "beamSearch2.cpp"
#include "zdd/convertEdgePermutation.cpp"
#include "zdd/weighted_sizeZdd.cpp"
#include "zdd/directed_degreeZdd.cpp"
#include "zdd/directed_relaxed_degreeZdd.cpp"
#include "zdd/pathZdd.cpp"
#include "zdd/pathZdd_char.cpp"
#include "util/width.cpp"

namespace zddSolver {

void solve(DiGraph G, std::vector<int>& perm, bool useMP) {
    // clock_t start;

    /* construct zdds */
           DirectedDegreeZDD         degree(G, perm);
    DirectedRelaxedDegreeZDD relaxed_degree(G, perm);
    WeightedSizeZDD size(G, perm);

    /* size restriction */
    // start = clock();
    tdzdd::DdStructure<2> dd(size, useMP);
    // auto node = dd.size(); 
    dd.zddReduce();
    // std::cout << node << " " << std::fixed << std::setprecision(4) << (double)(clock() - start) / CLOCKS_PER_SEC << " " << std::flush;
    
    /* relaxed degree restriction */
    // start = clock();
    dd.zddSubset(relaxed_degree);
    // node = dd.size();
    dd.zddReduce();
    // std::cout << node << " " << std::fixed << std::setprecision(4) << (double)(clock() - start) / CLOCKS_PER_SEC << " " << std::flush;
    
    /* degree restriction */
    // start = clock();
    dd.zddSubset(degree);
    // node = dd.size();
    dd.zddReduce();
    // std::cout << node << " " << std::fixed << std::setprecision(4) << (double)(clock() - start) / CLOCKS_PER_SEC << " " << std::flush;

    /* connectivity restriction */
    if(G.numVertices() < 255) {
        // start = clock();
        PathZDD_char path(G, perm);
        dd.zddSubset(path);
        // node = dd.size();
        // std::cout << node << " " << std::fixed << std::setprecision(4) << (double)(clock() - start) / CLOCKS_PER_SEC << std::endl << std::flush;
    } else {
        // start = clock();
        PathZDD path(G, perm);
        dd.zddSubset(path);
        // node = dd.size();
        // std::cout << node << " " << std::fixed << std::setprecision(4) << (double)(clock() - start) / CLOCKS_PER_SEC << std::endl << std::flush;
    }

    std::cout << dd.zddCardinality() << std::endl;
    // std::cout << dd.zddCardinality() << " ";
}

}
