#pragma once

#include "graph.cpp"
#include "digraph.cpp"

Graph readGraph() {
    int n, m, l, s, t;
    s = -1;
    t = -1;
    std::vector<std::pair<std::pair<int, int>, int>> edge;
    char buf[1024];
    while(fgets(buf, sizeof(buf), stdin)) {
        if(buf[0] == 'p') {
            sscanf(buf, "p edge %d %d", &n, &m);
        }
        else if(buf[0] == 'e') {
            int u, v;
            sscanf(buf, "e %d %d", &u, &v);
            u--;
            v--;
            std::pair<int, int> e = std::make_pair(u, v);
            edge.push_back(std::make_pair(e, 1));
        }
        /* [debug] */
        else if(buf[0] == 'a') {
            int u, v;
            sscanf(buf, "a %d %d", &u, &v);
            u--;
            v--;
            std::pair<int, int> e = std::make_pair(u, v);
            edge.push_back(std::make_pair(e, 1));
        }
        else if(buf[0] == 'l') {
            sscanf(buf, "l %d", &l);
        }
        else if(buf[0] == 't') {
            sscanf(buf, "t %d %d", &s, &t);
            s--;
            t--;
        }
        else if(buf[0] == 'c') {
            continue;
        }
    }
    std::sort(edge.begin(), edge.end());
    Graph G(n, m, l, s, t, edge);
    return G;
}

/*
Graph readGraph(const char *file) {
    int n, m, l, s, t;
    s = -1;
    t = -1;
    std::vector<std::pair<std::pair<int, int>, int>> edge;
    char buf[1024];
    FILE *fp;
    if((fp = fopen(file, "r")) == NULL) {
        fprintf(stderr, "Error: cannot open file\n");
        exit(1);
    }
    while(fgets(buf, sizeof(buf), fp)) {
        if(buf[0] == 'p') {
            sscanf(buf, "p edge %d %d", &n, &m);
        }
        else if(buf[0] == 'e') {
            int u, v;
            sscanf(buf, "e %d %d", &u, &v);
            u--;
            v--;
            std::pair<int, int> e = std::make_pair(u, v);
            edge.push_back(std::make_pair(e, 1));
        }
        else if(buf[0] == 'l') {
            sscanf(buf, "l %d", &l);
        }
        else if(buf[0] == 't') {
            sscanf(buf, "t %d %d", &s, &t);
            s--;
            t--;
        }
        else if(buf[0] == 'c') {
            continue;
        }
    }
    std::sort(edge.begin(), edge.end());
    Graph G(n, m, l, s, t, edge);
    return G;
}
*/

void writeGraph(Graph G, const char *file) {
    int n = G.numVertices();
    int m = G.numEdges();
    int l = G.numLength();
    int s = G.getStart();
    int t = G.getTerminal();
    s++;
    t++;
    FILE *fp = fopen(file, "wt");
    fprintf(fp, "p edge %d %d\n", n, m);
    for(int i = 0; i < m; i++) {
        int u = G.getEdge(i).first.first;
        int v = G.getEdge(i).first.second;
        int w = G.getEdge(i).second;
        u++;
        v++;
        fprintf(fp, "e %d %d\n", u, v);
    }
    fprintf(fp, "l %d\n", l);
    if(G.isOnepair()) fprintf(fp, "t %d %d\n", s, t);
    fclose(fp);
}

void writeWeightedGraph(Graph G, const char *file) {
    int n = G.numVertices();
    int m = G.numEdges();
    int l = G.numLength();
    int s = G.getStart();
    int t = G.getTerminal();
    s++;
    t++;
    FILE *fp = fopen(file, "wt");
    fprintf(fp, "p edge %d %d\n", n, m);
    for(int i = 0; i < m; i++) {
        int u = G.getEdge(i).first.first;
        int v = G.getEdge(i).first.second;
        int w = G.getEdge(i).second;
        u++;
        v++;
        fprintf(fp, "e %d %d %d\n", u, v, w);
    }
    fprintf(fp, "l %d\n", l);
    if(G.isOnepair()) fprintf(fp, "t %d %d\n", s, t);
    fclose(fp);
}

// void writeGraphWithColor(Graph G, const char *file, std::vector<int> comp) {
//     int n = G.numVertices();
//     int m = G.numEdges();
//     int l = G.numLength();
//     int s = G.getStart();
//     int t = G.getTerminal();
//     s++;
//     t++;
//     FILE *fp = fopen(file, "wt"), *pp;
//     fprintf(fp, "p edge %d %d\n", n, m);
//     for(int i = 0; i < n; i++) {
//         fprintf(fp, "v %d %d\n", i + 1, comp[i]);
//     }
//     for(int i = 0; i < m; i++) {
//         int u = G.getEdge(i).first.first;
//         int v = G.getEdge(i).first.second;
//         int w = G.getEdge(i).second;
//         u++;
//         v++;
//         fprintf(fp, "e %d %d\n", u, v);
//     }
//     fprintf(fp, "l %d\n", l);
//     if(G.isOnepair()) fprintf(fp, "t %d %d\n", s, t);
//     fclose(fp);

//     /* python script */
//     char proc[1024];
//     sprintf(proc, "python3 instances-draw/draw_color.py %s", file);
//     if((pp = popen(proc, "r")) == NULL) {
//         fprintf(stderr, "Error: popen()\n");
//         return ;
//     }
//     if(pclose(pp) == -1) {
//         fprintf(stderr, "Error: pclose()\n");
//         return ;
//     }
// }

DiGraph readDiGraph() {
    int n, m, l, s, t;
    s = -1;
    t = -1;
    std::vector<std::pair<std::pair<int, int>, int>> edge;
    char buf[1024];
    while(fgets(buf, sizeof(buf), stdin)) {
        if(buf[0] == 'p') {
            sscanf(buf, "p edge %d %d", &n, &m);
        }
        else if(buf[0] == 'e') {
            int u, v;
            sscanf(buf, "e %d %d", &u, &v);
            u--;
            v--;
            std::pair<int, int> e = std::make_pair(u, v);
            edge.push_back(std::make_pair(e, 1));
            std::pair<int, int> f = std::make_pair(v, u);
            edge.push_back(std::make_pair(f, 1));
        }
        else if(buf[0] == 'a') {
            int u, v;
            sscanf(buf, "a %d %d", &u, &v);
            u--;
            v--;
            std::pair<int, int> e = std::make_pair(u, v);
            edge.push_back(std::make_pair(e, 1));
        }
        else if(buf[0] == 'l') {
            sscanf(buf, "l %d", &l);
        }
        else if(buf[0] == 't') {
            sscanf(buf, "t %d %d", &s, &t);
            s--;
            t--;
        }
        else if(buf[0] == 'c') {
            continue;
        }
    }
    std::sort(edge.begin(), edge.end());
    DiGraph G(n, m, l, s, t, edge);
    return G;
}

/*
DiGraph readDiGraph(const char *file) {
    int n, m, l, s, t;
    s = -1;
    t = -1;
    std::vector<std::pair<std::pair<int, int>, int>> edge;
    char buf[1024];
    FILE *fp;
    if((fp = fopen(file, "r")) == NULL) {
        fprintf(stderr, "Error: cannot open file\n");
        exit(1);
    }
    while(fgets(buf, sizeof(buf), fp)) {
        if(buf[0] == 'p') {
            sscanf(buf, "p edge %d %d", &n, &m);
        }
        else if(buf[0] == 'e') {
            int u, v;
            sscanf(buf, "e %d %d", &u, &v);
            u--;
            v--;
            std::pair<int, int> e = std::make_pair(u, v);
            edge.push_back(std::make_pair(e, 1));
            std::pair<int, int> f = std::make_pair(v, u);
            edge.push_back(std::make_pair(f, 1));
        }
        else if(buf[0] == 'a') {
            int u, v;
            sscanf(buf, "a %d %d", &u, &v);
            u--;
            v--;
            std::pair<int, int> e = std::make_pair(u, v);
            edge.push_back(std::make_pair(e, 1));
        }
        else if(buf[0] == 'l') {
            sscanf(buf, "l %d", &l);
        }
        else if(buf[0] == 't') {
            sscanf(buf, "t %d %d", &s, &t);
            s--;
            t--;
        }
        else if(buf[0] == 'c') {
            continue;
        }
    }
    std::sort(edge.begin(), edge.end());
    DiGraph G(n, m, l, s, t, edge);
    return G;
}
*/

void writeDiGraph(DiGraph G, const char *flie) {
    int n = G.numVertices();
    int m = G.numEdges();
    int l = G.numLength();
    int s = G.getStart();
    int t = G.getTerminal();
    s++;
    t++;
    FILE *fp = fopen(flie, "wt");
    fprintf(fp, "p edge %d %d\n", n, m);
    for(int i = 0; i < m; i++) {
        int u = G.getEdge(i).first.first;
        int v = G.getEdge(i).first.second;
        int w = G.getEdge(i).second;
        u++;
        v++;
        fprintf(fp, "a %d %d\n", u, v);
    }
    fprintf(fp, "l %d\n", l);
    if(G.isOnepair()) fprintf(fp, "t %d %d\n", s, t);
    fclose(fp);
}

void writeWeightedDiGraph(DiGraph G, const char *flie) {
    int n = G.numVertices();
    int m = G.numEdges();
    int l = G.numLength();
    int s = G.getStart();
    int t = G.getTerminal();
    s++;
    t++;
    FILE *fp = fopen(flie, "wt");
    fprintf(fp, "p edge %d %d\n", n, m);
    for(int i = 0; i < m; i++) {
        int u = G.getEdge(i).first.first;
        int v = G.getEdge(i).first.second;
        int w = G.getEdge(i).second;
        u++;
        v++;
        fprintf(fp, "a %d %d %d\n", u, v, w);
    }
    fprintf(fp, "l %d\n", l);
    if(G.isOnepair()) fprintf(fp, "t %d %d\n", s, t);
    fclose(fp);
}
