#include <iostream>
#include <vector>
#include <algorithm>

#include "graphio.cpp"
#include "delete_vertices.cpp"
#include "zddSolver/solve.cpp"

Graph preprocessGraph(const Graph& graph) {
    std::vector<std::pair<int, int>> edge(graph.numEdges());

    for (int i = 0; i < graph.numEdges(); ++i) {
        std::pair<std::pair<int, int>, int> e = graph.getEdge(i);
        edge[i] = e.first;
    }

    tachibana::graph tmpgraph(graph.numVertices(), edge, graph.getStart(), graph.getTerminal(), graph.numLength(), false);
    tachibana::preprocess pr(tmpgraph);
    pr.deleteVertices();
    int n = pr.getNumVertices(), s = pr.getStart(), t = pr.getTerminal(), limit = pr.getLength();
    std::vector<std::pair<std::pair<int, int>, int>> e(pr.getNumEdges());
    for (int i = 0; i < pr.getNumEdges(); ++i) {
        e[i] = {pr.getEdge(i), 1};
    }
    Graph preprog(n, pr.getNumEdges(), limit, s, t, e);
    Graph hg = preprog.deleteVertices().deleteLeaves().deletePaths();

    // writeGraph(hg, "out.txt");
    return hg;
}

int main(int argc, char **argv) {
    bool useMP = true;

    /* preprocessing */
    Graph G = preprocessGraph(readGraph());

    /* path decomposition */
    std::vector<int> perm;
    if(G.numVertices() < 4000) {
        auto bs_node = BS::beamSearch(400, 40000, G);
        perm = BS::getVertexPermutation(bs_node);
    } else {
        auto bs_node = BS2::beamSearch(1, 10000, G);
        perm = BS2::getVertexPermutation(bs_node);
    }

    /* solve */
    zddSolver::solve(G, perm, useMP);
}
