#pragma once

#include <time.h>
#include <string>

#include "../graph.cpp"
#include "../graphio.cpp"
#include "beamSearch.cpp"
#include "beamSearch2.cpp"
#include "zdd/convertEdgePermutation.cpp"
#include "zdd/weighted_sizeZdd.cpp"
#include "zdd/degreeZdd.cpp"
#include "zdd/degreeZdd_sep.cpp"
#include "zdd/relaxed_degreeZdd.cpp"
#include "zdd/relaxed_degreeZdd_sep.cpp"
#include "zdd/pathZdd.cpp"
#include "zdd/pathZdd_char.cpp"
#include "util/width.cpp"

namespace zddSolver {

void solve(Graph G, std::vector<int> perm, bool useMP) {
    /* construct zdds */
           DegreeZDD         degree(G, perm);
    RelaxedDegreeZDD relaxed_degree(G, perm);
    WeightedSizeZDD size(G, perm);

    /* size restriction */
    tdzdd::DdStructure<2> dd(size, useMP);
    dd.zddReduce();
    
    /* relaxed degree restriction */
    dd.zddSubset(relaxed_degree);
    dd.zddReduce();
    
    /* degree restriction */
    dd.zddSubset(degree);
    dd.zddReduce();

    /* connectivity restriction */
    if(G.numVertices() < 255) {
        PathZDD_char path(G, perm);
        dd.zddSubset(path);
    } else {
        PathZDD path(G, perm);
        dd.zddSubset(path);
    }

    std::cout << dd.zddCardinality() << std::endl;
}

}
