// 強連結成分を用いた数え上げ
// 数え上げの関数はcalcpathnumberです
#pragma once

#include <iostream>
#include <vector>
#include <algorithm>

#include <boost/multiprecision/cpp_int.hpp>
// #define boost/multiprecision/cpp_int.hpp int;

#include "../graph.cpp"
#include "../digraph.cpp"
#include "../graphio.cpp"
#include "../graph_bfs.cpp"

#include "pathcounting_func.cpp"

#include "time.h"

namespace SCCpathcount
{
    // 型名(仮)
    using hoge = long long;

    // 強連結成分class
    class SCC
    {
        std::vector<int> starts, terminals, vertexID;
        int numvertex, limit;
        std::vector<std::pair<std::pair<int, int>, int>> edges;
    public:
        SCC(int l = 1 << 30) : limit(l)
        {
            starts.clear();
            terminals.clear();
            vertexID.clear();
            edges.clear();
        }

        void add_edge(const int u, const int v, int w = 1) {edges.push_back({{u, v}, w});}
        void add_start(const int v) {starts.emplace_back(v);}
        void add_terminal(const int v) {terminals.emplace_back(v);}

        void print()
        {
            std::cerr << "print: |starts|=" << starts.size() << ", |terminals|=" << terminals.size() << std::endl;
            for (int v : starts) std::cerr << ' ' << v;
            std::cerr << std::endl;
            for (int v : terminals) std::cerr << ' ' << v;
            std::cerr << std::endl;
        }

        void build()
        {
            vertexID.resize(edges.size() * 2 + starts.size() + terminals.size());
            for (int i = 0; i < edges.size(); ++i)
            {
                vertexID[i + edges.size()] = edges[i].first.first;
                vertexID[i] = edges[i].first.second;
            }
            for (int i = 0; i < starts.size(); ++i) vertexID[i + edges.size() * 2] = starts[i];
            for (int i = 0; i < terminals.size(); ++i) vertexID[i + edges.size() * 2 + starts.size()] = terminals[i];


            std::sort(vertexID.begin(), vertexID.end());
            vertexID.erase(std::unique(vertexID.begin(), vertexID.end()), vertexID.end());
            numvertex = vertexID.size();
            for (auto &[uv, w] : edges)
            {
                auto &[u, v] = uv;
                int uid = std::lower_bound(vertexID.begin(), vertexID.end(), u) - vertexID.begin();
                int vid = std::lower_bound(vertexID.begin(), vertexID.end(), v) - vertexID.begin();
                u = uid;
                v = vid;
            }
            std::sort(starts.begin(), starts.end());
            std::sort(terminals.begin(), terminals.end());
            starts.erase(std::unique(starts.begin(), starts.end()), starts.end());
            terminals.erase(std::unique(terminals.begin(), terminals.end()), terminals.end());
            for (auto &v : starts)
            {
                int vid = std::lower_bound(vertexID.begin(), vertexID.end(), v) - vertexID.begin();
                v = vid;
            }
            for (auto &v : terminals)
            {
                int vid = std::lower_bound(vertexID.begin(), vertexID.end(), v) - vertexID.begin();
                v = vid;
            }
            int wsum = 0;
            std::vector<int> maxweight(numvertex, 1);
            for (auto &[uv, w] : edges)
            {
                auto [u, v] = uv;
                maxweight[u] = std::max(maxweight[u], w);
            }
            for (int i = 0; i < numvertex; ++i) wsum += maxweight[i];
            limit = std::min(limit, wsum);
        }

        void setLimit(const int l) {limit = l;}

        int getNumStart() {return starts.size();}
        int getStart(int id) {return vertexID[starts[id]];}
        int getNumTerminal() {return terminals.size();}
        int getTerminal(int id) {return vertexID[terminals[id]];}
        int getLimit() {return limit;}

        // ここの実装をお願いします
        std::vector<std::vector<std::vector<hoge>>> calcpathnumber()
        {
            // cnt[i][j][k] := 辺集合eで構成される有向グラフにおいて，長さが k の starts[i] -> terminals[j] パスの数
            // 型名<hoge>の部分は std::string, long long などにしてください
            // std::vector<std::vector<std::vector<hoge>>> cnt(starts.size(), std::vector<std::vector<hoge>> (terminals.size(), std::vector<hoge> (limit + 1)));
            // std::vector<std::vector<std::pair<int, int>>> g(numvertex);
            // std::vector<std::vector<std::vector<hoge>>> tmp(numvertex, std::vector<std::vector<hoge>> (numvertex, std::vector<hoge> (limit + 1)));
            // for (auto &[uv, w] : edges)
            // {
            //     auto [u, v] = uv;
            //     g[u].emplace_back(v, w);
            // }
            // std::vector<int> check(numvertex);
            // int p = 0;
            // auto dfs = [&](auto f, int now, int len) -> void
            // {
            //     if (len > limit || check[now]) return;
            //     check[now] = 1;
            //     ++tmp[p][now][len];
            //     for (auto [to, w] : g[now])
            //     {
            //         f(f, to, len + w);
            //     }
            //     check[now] = 0;
            // };
            // for (p = 0; p < numvertex; ++p) dfs(dfs, p, 0);

            // for (int i = 0; i < starts.size(); ++i)
            // {
            //     for (int j = 0; j < terminals.size(); ++j)
            //     {
            //         std::swap(tmp[starts[i]][terminals[j]], cnt[i][j]);
            //     }
            // }

            return sccCounting(numvertex, edges, starts, terminals, limit);
        }
    };

    std::vector<SCC> scc;
    std::vector<int> groupid;

    // 強連結成分分解(各頂点のグループ番号を求める)
    int SCCdecompose(int n, std::vector<std::pair<std::pair<int, int>, int>> &e)
    {
        std::vector<std::vector<int>> g(n), grev(n);
        groupid.assign(n, 0);
        for (auto &[uv, w] : e)
        {
            auto &[u, v] = uv;
            g[u].emplace_back(v);
            grev[v].emplace_back(u);
        }
        std::vector<int> perm;
        auto topsort = [&](auto f, int now) -> void
        {
            if (groupid[now]) return;
            groupid[now] = -1;
            for (auto v : g[now]) f(f, v);
            perm.push_back(now);
        };
        for (int i = 0; i < n; ++i) topsort(topsort, i);
        reverse(perm.begin(), perm.end());
        int grsiz = 0;
        auto calgrid = [&](auto f, int now) -> void
        {
            if (groupid[now] != -1) return;
            groupid[now] = grsiz;
            for (auto v : grev[now]) f(f, v);
        };
        for (auto v : perm)
        {
            if (groupid[v] != -1) continue;
            calgrid(calgrid, v);
            ++grsiz;
        }
        scc.resize(grsiz);
        return grsiz;
    }

    std::string SCCpathcount(DiGraph &g)
    {
        int n = g.numVertices(), m = g.numEdges(), s = g.getStart(), t = g.getTerminal(), lim = g.numLength();
        std::vector<std::pair<std::pair<int, int>, int>> edge(m);
        for (int i = 0; i < m; ++i) edge[i] = g.getEdge(i);
        int groupnum = SCCdecompose(n, edge);
        scc.resize(groupnum);

        std::vector<std::vector<std::pair<int, int>>> sccgraph(n);
        std::vector<std::pair<std::pair<int, int>, int>> groupedge;
        for (auto [uv, w] : edge)
        {
            auto [u, v] = uv;
            if (groupid[u] == groupid[v])
            {
                scc[groupid[u]].add_edge(u, v, w);
            }
            else
            {
                scc[groupid[u]].add_terminal(u);
                scc[groupid[v]].add_start(v);
                sccgraph[u].emplace_back(v, w);
            }
        }
        scc[groupid[s]].add_start(s);
        scc[groupid[t]].add_terminal(t);

        for (int i = 0; i < groupnum; ++i) scc[i].setLimit(lim);
        for (int i = 0; i < groupnum; ++i){
            scc[i].build();
        }

        std::vector<std::vector<boost::multiprecision::cpp_int>> dp(n, std::vector<boost::multiprecision::cpp_int> (lim + 1));
        dp[s][0] = 1;
        for (int i = 0; i < groupnum; ++i)
        {
            auto res = scc[i].calcpathnumber();
            int ssiz = scc[i].getNumStart(), tsiz = scc[i].getNumTerminal();
            std::vector<std::vector<boost::multiprecision::cpp_int>> ndp(tsiz, std::vector<boost::multiprecision::cpp_int> (lim + 1));
            for (int k = 0; k < tsiz; ++k)
            {
                int teid = scc[i].getTerminal(k);
                for (int j = 0; j < ssiz; ++j)
                {
                    int stid = scc[i].getStart(j);
                    for (int l1 = 0; l1 <= scc[i].getLimit(); ++l1)
                    {
                        boost::multiprecision::cpp_int cnt1(res[j][k][l1]);
                        for (int l2 = 0; l2 + l1 <= lim; ++l2)
                        {
                            ndp[k][l1 + l2] += dp[stid][l2] * cnt1;
                        }
                    }
                }
            }
            for (int j = 0; j < tsiz; ++j) for (int k = 0; k <= lim; ++k) dp[scc[i].getTerminal(j)][k] = ndp[j][k];
            for (int j = 0; j < tsiz; ++j)
            {
                int teid = scc[i].getTerminal(j);
                for (auto &[to, w] : sccgraph[teid])
                {
                    for (int l = 0; l + w <= lim; ++l)
                    {
                        dp[to][l + w] += dp[teid][l];
                    }
                }
            }
        }
        boost::multiprecision::cpp_int pathcount = 0;
        for (int i = 0; i <= lim; ++i) pathcount += dp[t][i];
        return pathcount.str();
    }
}

namespace hamSolver {

void solve(const DiGraph& graph) {
    clock_t start = clock();

    DiGraph graph_bfs = pathBfs(graph, graph.getStart());
    std::string ans = SCCpathcount::SCCpathcount(graph_bfs);
    
    std::cout << ans ;//<< std::endl; 

    // clock_t end = clock();
    // std::cout << "," << double(end-start) / CLOCKS_PER_SEC << std::endl;
}

}

