#include <iostream>
#include <vector>
#include <algorithm>

#include "graphio.cpp"
#include "delete_vertices.cpp"
#include "zddSolver/solve.cpp"
#include "hamSolver/SCCpathcounting.cpp"

DiGraph preprocessGraph(const DiGraph& graph) {
    std::vector<std::pair<int, int>> edge(graph.numEdges());

    for (int i = 0; i < graph.numEdges(); ++i) {
        std::pair<std::pair<int, int>, int> e = graph.getEdge(i);
        edge[i] = e.first;
    }

    tachibana::graph tmpgraph(graph.numVertices(), edge, graph.getStart(), graph.getTerminal(), graph.numLength(), true);
    tachibana::preprocess pr(tmpgraph);
    pr.deleteVertices();
    int n = pr.getNumVertices(), s = pr.getStart(), t = pr.getTerminal(), limit = pr.getLength();
    std::vector<std::pair<std::pair<int, int>, int>> e(pr.getNumEdges());
    for (int i = 0; i < pr.getNumEdges(); ++i) {
        e[i] = {pr.getEdge(i), 1};
    }
    DiGraph preprog(n, pr.getNumEdges(), limit, s, t, e);
    DiGraph hg = preprog; //.contraction2();

    // writeGraph(hg, "out.txt");
    return hg;
}

int main(int argc, char **argv) {
    bool useMP = false;

    /* preprocessing */
    DiGraph G = preprocessGraph(readDiGraph());
    // DiGraph G = readDiGraph();

    DiGraph H = G.contraction2();
    /* path decomposition */
    std::vector<int> perm;
    Graph Gu = H.convertUndirectedGraph();
    if(Gu.numVertices() < 4000) {
        auto bs_node = BS::beamSearch(400, 40000.0, Gu);
        perm = BS::getVertexPermutation(bs_node);
    } else {
        auto bs_node = BS2::beamSearch(1, 10000.0, Gu);
        perm = BS2::getVertexPermutation(bs_node);
    }

    /* solve */
    if (G.numLength() > 10 && pathWidth(H, perm) < 20) {
        zddSolver::solve(H, perm, useMP);
    }
    else {
        hamSolver::solve(G);
    }
    
}