#pragma once

#include <iostream>
#include <fstream>
#include <vector>
#include <set>
#include <map>
#include <algorithm>

#include "../../digraph.cpp"

std::vector<std::pair<int, int>> convertEdgePermutation_directed(DiGraph &G, std::vector<int> &perm) {
    int n = G.numVertices();
    int m = G.numEdges();
    assert(n == perm.size());
    std::vector<std::vector<int>> graph(n);
    std::map<std::pair<int, int>, std::vector<int>> edge_weight;
    for(int i = 0; i < m; i++) {
        int u = G.getEdge(i).first.first;
        int v = G.getEdge(i).first.second;
        int c = G.getEdge(i).second;
        if(edge_weight.count(std::make_pair(u, v)) || 
           edge_weight.count(std::make_pair(v, u))) {
            edge_weight[std::make_pair(u, v)].push_back(c);
        } else {
            graph[u].push_back(v);
            graph[v].push_back(u);
            edge_weight[std::make_pair(u, v)].push_back(c);
        }
    }

    std::vector<int> invPerm(n);
    for(int i = 0; i < n; i++) invPerm[perm[i]] = i;
    std::vector<int> is_frontier(n);

    std::vector<bool> used(n);
    std::vector<std::pair<int, int>> res;
    for(auto i: perm) {
        std::vector<int> tmp;
        for(auto j: graph[i]) {
            if(!used[j]) tmp.push_back(j);
        }

        sort(tmp.begin(), tmp.end(), [&](int a, int b) -> bool {
            if(is_frontier[a] == is_frontier[b]) {
                return invPerm[a] < invPerm[b];
            }
            return is_frontier[a] > is_frontier[b];
        });
        for(auto j: tmp) {
            {
                std::pair<int, int> e = std::make_pair(i, j);
                for(int k = 0; k < edge_weight[e].size(); k++) res.push_back(e);
            }
            {
                std::pair<int, int> e = std::make_pair(j, i);
                for(int k = 0; k < edge_weight[e].size(); k++) res.push_back(e);
            }
            is_frontier[j] = true;
        }
        used[i] = true;
    }
    return res;
}

std::vector<int> convertEdgePermutation_weighted_directed(DiGraph &G, std::vector<int> &perm) {
    int n = G.numVertices();
    int m = G.numEdges();
    assert(n == perm.size());
    std::vector<std::vector<int>> graph(n);
    std::map<std::pair<int, int>, std::vector<int>> edge_weight;
    for(int i = 0; i < m; i++) {
        int u = G.getEdge(i).first.first;
        int v = G.getEdge(i).first.second;
        int c = G.getEdge(i).second;
        if(edge_weight.count(std::make_pair(u, v)) || 
           edge_weight.count(std::make_pair(v, u))) {
            edge_weight[std::make_pair(u, v)].push_back(c);
        } else {
            graph[u].push_back(v);
            graph[v].push_back(u);
            edge_weight[std::make_pair(u, v)].push_back(c);
        }
    }

    std::vector<int> invPerm(n);
    for(int i = 0; i < n; i++) invPerm[perm[i]] = i;
    std::vector<int> is_frontier(n);

    std::vector<bool> used(n);
    std::vector<int> res;
    for(auto i: perm) {
        std::vector<int> tmp;
        for(auto j: graph[i]) {
            if(!used[j]) tmp.push_back(j);
        }

        sort(tmp.begin(), tmp.end(), [&](int a, int b) -> bool {
            if(is_frontier[a] == is_frontier[b]) {
                return invPerm[a] < invPerm[b];
            }
            return is_frontier[a] > is_frontier[b];
        });
        for(auto j: tmp) {
            {
                std::pair<int, int> e = std::make_pair(i, j);
                for(int k = 0; k < edge_weight[e].size(); k++) res.push_back(edge_weight[e][k]);
            }
            {
                std::pair<int, int> e = std::make_pair(j, i);
                for(int k = 0; k < edge_weight[e].size(); k++) res.push_back(edge_weight[e][k]);
            }
            is_frontier[j] = true;
        }
        used[i] = true;
    }
    return res;
}
