#include <ZBDD.h>
#include "CtoI.h" 
#include "SBDD_helper.h"
using namespace sbddh;
#include <iostream> // for test output
#include <vector>
#ifdef MAP
#include <map>
std::map<std::pair<bddword, bddword>, ZBDD> cacheDJ; // Cache for Disjoint join
std::map<std::pair<bddword, bddword>, ZBDD> cacheJJ; // Cache for Joint Join
#else
/*
#include <unordered_map> // c++11はTdZddがコンパイルできず?
class Hash_UM{
public:
  size_t operator()(const std::pair<bddword, bddword> p) const{
    return 3*p.first+7*p.second;
  }
};
std::unordered_map<std::pair<bddword, bddword>, ZBDD, Hash_UM> cacheDJ; // Cache for Disjoint join
std::unordered_map<std::pair<bddword, bddword>, ZBDD, Hash_UM> cacheJJ; // Cache for Joint Join
*/
#include <boost/unordered_map.hpp>

class Hash_UM{
public:
  size_t operator()(const std::pair<bddword, bddword> p) const{
    return 997*p.first+p.second;
  }
};
boost::unordered_map<std::pair<bddword, bddword>, ZBDD, Hash_UM> cacheDJ; // Cache for Disjoint join
boost::unordered_map<std::pair<bddword, bddword>, ZBDD, Hash_UM> cacheJJ; // Cache for Joint Join
#endif

ZBDD exorZdd(ZBDD x, ZBDD y) {
    ZBDD z;
    z = (x + y) - (x & y);
    return z;
}

ZBDDV exorVecZdd(ZBDDV x, ZBDDV y) {
    ZBDDV z;

    return z;
}

std::vector<ZBDD> exorVecZdd(std::vector<ZBDD> x, std::vector<ZBDD> y) {
    std::vector<ZBDD> z;
    ZBDD C;

    int n = std::min(x.size(), y.size());
    for (int i = 0; i < n; i++) {
        if (i == 0) { // z[0]
            z.push_back(exorZdd(x[i], y[i]));
            C = x[i] & y[i];
        } else {  // z[i] : carry up from [i-1](?)
            z.push_back(exorZdd(exorZdd(x[i], y[i]), C));
            C = (x[i] & y[i]) + (x[i] & C) + (y[i] & C);
        }
    }
    if(x.size() < y.size()){
        for(int i = n ; i<y.size() ; i++){
            z.push_back(exorZdd(y[i], C));
            C = y[i] & C;
        }
    }
    else{
        for(int i = n ; i<x.size() ; i++){
            z.push_back(exorZdd(x[i], C));
            C = x[i] & C;
        }
    }
    if(C != ZBDD(0)) z.push_back(C);
    return z;
}

std::vector<ZBDD> multiple_term1(std::vector<ZBDD> f, std::vector<ZBDD> g) {
  // 1終端節点に行き着いた組合せC1,C2の数を掛け合わせる
  int fs = f.size(), gs = g.size();
  std::vector<ZBDD> re;
  re.push_back(ZBDD(0));

  for (int i = 0; i < gs; i++) {
    std::vector<ZBDD> ij;
    for (int j = 0; j < fs; j++) {
      ij.push_back(f[j]&g[i]);
    }

    for (int k = 0; k < i; k++) {
        ij.insert(ij.begin(), ZBDD(0));
    }

    re = exorVecZdd(re, ij);
  }

  // diet
  while(re[re.size()-1] == ZBDD(0)) {
    re.erase(re.end() - 1);
  }
  return re;
}

// f $ g = a ( f_0 $ g_1 ∪ f_1 $ g_0) ∪ (f_0 $ g_0)
// T $ T = T
// T $ F = F
// F $ T = F
// F $ F = F
ZBDD disjointUnion(ZBDD f, ZBDD g, int dep){
  std::cout << dep << "-f: " << ZStr(f) << std::endl;
  std::cout << dep << "-g: " << ZStr(g) << std::endl;

  //std::cerr << ftop-20 << ", " << gtop-20 << std::endl;
  
  if(f == ZBDD(1) && g == ZBDD(1)) {std::cout << " return 1: " << ZStr(ZBDD(1)) << std::endl; return ZBDD(1);}
  if(f == ZBDD(0) || g == ZBDD(0)) {std::cout << " return 0: " << ZStr(ZBDD(0)) << std::endl; return ZBDD(0);}

  if(f.GetID() > g.GetID()){
    ZBDD temp;
    temp = f;
    f = g;
    g = temp;
  }
#ifndef NON_CACHE
  std::pair<bddword, bddword> key = std::pair<bddword, bddword> (f.GetID(), g.GetID());
  if(cacheDJ.find(key) != cacheDJ.end()) {std::cout << dep << "-k: " << ZStr(cacheDJ[key]) << std::endl; return cacheDJ[key];}
#endif

  int ftop = f.Top();
  int gtop = g.Top();
  
  ZBDD f0, f1, g0, g1;
  if(ftop != 0){
    f0 = f.OffSet(ftop);
    f1 = f.OnSet0(ftop);
  }
  if(gtop != 0){
    g0 = g.OffSet(gtop);
    g1 = g.OnSet0(gtop);
  }

  ZBDD z;
  if(ftop < gtop) z = disjointUnion(f, g1, dep+1).Change(gtop) + disjointUnion(f, g0, dep+1);
  if(gtop < ftop) z = disjointUnion(g, f1, dep+1).Change(ftop) + disjointUnion(g, f0, dep+1);
  if(ftop == gtop) z = (disjointUnion(f0, g1, dep+1) + disjointUnion(f1, g0, dep+1)).Change(ftop) + disjointUnion(f0, g0, dep+1);

#ifndef NON_CACHE
  cacheDJ.insert(std::make_pair(key, z));
#endif
  std::cout << dep << "-z: " << ZStr(z) << std::endl;
  return z;
}

boost::unordered_map<std::pair<bddword, bddword>, std::vector<ZBDD>, Hash_UM> cacheDJv; // Cache for Disjoint join
std::vector<ZBDD> disjointUnionVec(std::vector<ZBDD> f, std::vector<ZBDD> g){
  
  // check input
  int ftop0 = 0, gtop0 = 0;
  for (auto x: f) {
    if (x.Top() > ftop0) {
      ftop0 = x.Top();
    }
  }
  for (auto x: g) {
    if (x.Top() > gtop0) {
      gtop0 = x.Top();
    }
  }
  if((ftop0 == 0 && f[0] == ZBDD(1)) && (gtop0 == 0 && g[0] == ZBDD(1))) {
    std::vector<ZBDD> re; re.push_back(ZBDD(1)); 
    return re;
  }
  if(ftop0 == 0 && f[0] == ZBDD(0)) {
    std::vector<ZBDD> re; re.push_back(ZBDD(0));  
    return re;
  }
  if(gtop0 == 0 && g[0] == ZBDD(0)) {
    std::vector<ZBDD> re; re.push_back(ZBDD(0)); 
    return re;
  }
  //=========
  // chek key
  bddword fidmax = 0, gidmax = 0;
  int fidp = 0, gidp = 0;
  for (int i = 0; i < f.size(); i++) {
    if (f[i].GetID() > fidmax) {
      fidmax = f[i].GetID();
      fidp = i;
    }
  }
  for (int i = 0; i < g.size(); i++) {
    if (g[i].GetID() > gidmax) {
      gidmax = g[i].GetID();
      gidp = i;
    }
  }
  if(fidmax > gidmax){
    std::vector<ZBDD> temp;
    bddword idtemp;
    temp = f;
    f = g;
    g = temp;
    idtemp = fidmax;
    fidmax = gidmax;
    gidmax = idtemp;
  }
#ifndef NON_CACHE
  std::pair<bddword, bddword> key = std::pair<bddword, bddword> (fidmax, gidmax);
  if(cacheDJv.find(key) != cacheDJv.end()) {
    return cacheDJv[key];
  }
#endif

  // delete top
  int ftop = 0;
  int gtop = 0;
  for (auto x: f) {
    if (x.Top() > ftop) {
      ftop = x.Top();
    }
  }
  for (auto x: g) {
    if (x.Top() > gtop) {
      gtop = x.Top();
    }
  }
  
  std::vector<ZBDD> f0, f1, g0, g1;
  if(ftop != 0){
    for (auto x : f) {
      f0.push_back(x.OffSet(ftop));
      f1.push_back(x.OnSet0(ftop));
    }
  }
  if(gtop != 0){
    for (auto x : g) {
      g0.push_back(x.OffSet(gtop));
      g1.push_back(x.OnSet0(gtop));
    }
  }

  // making disjoint
  std::vector<ZBDD> z;
  if(ftop < gtop) {
    z = disjointUnionVec(f, g1);
    for (int i = 0; i < z.size(); i++) {
      z[i] = z[i].Change(gtop);
    }
    z = exorVecZdd(z, disjointUnionVec(f, g0));
  }
  if(gtop < ftop) {
    z = disjointUnionVec(g, f1);
    for (int i = 0; i < z.size(); i++) {
      z[i] = z[i].Change(ftop);
    }
    z = exorVecZdd(z, disjointUnionVec(g, f0));
  }

  if(ftop == gtop) {
    z = exorVecZdd(disjointUnionVec(f0, g1), disjointUnionVec(f1, g0));
    for (int i = 0; i < z.size(); i++) {
      z[i] = z[i].Change(ftop);
    }
    z = exorVecZdd(z, disjointUnionVec(f0, g0));
  };

#ifndef NON_CACHE
  cacheDJv.insert(std::make_pair(key, z));
#endif
  
  return z;
}

// no need multi and map ver.
//boost::unordered_map<std::pair<bddword, bddword>, std::vector<ZBDD>, Hash_UM> cacheDJv; // Cache for Disjoint join
std::vector<ZBDD> disjointUnionVec(std::vector<ZBDD> f, std::vector<ZBDD> g, int dep){

  
  // check input and id
  int ftop = 0, gtop = 0;
  ZBDD fid = ZBDD(0), gid = ZBDD(0);
  bddword fidmax = fid.GetID(), gidmax = gid.GetID();
  for (auto x: f) {
    if (x.Top() > ftop) {
        ftop = x.Top();
    }
    if (fidmax < x.GetID()) {
        fid = x;
    }
  }
  for (auto x: g) {
    if (x.Top() > gtop) {
        gtop = x.Top();
    }
    if (gidmax < x.GetID()) {
        gid = x;
    }
  }
  if((ftop == 0 && fid == ZBDD(1)) && (gtop == 0 && gid == ZBDD(1))) {
    std::vector<ZBDD> re;
    re = multiple_term1(f, g); 
    return re;
  }
  if(ftop == 0 && fid == ZBDD(0)) {
    std::vector<ZBDD> re; re.push_back(ZBDD(0));  
    return re;
  }
  if(gtop == 0 && gid == ZBDD(0)) {
    std::vector<ZBDD> re; re.push_back(ZBDD(0)); 
    return re;
  }
  //=========
  
  std::vector<ZBDD> f0, f1, g0, g1;
  if(ftop != 0){
    for (auto x : f) {
      f0.push_back(x.OffSet(ftop));
      f1.push_back(x.OnSet0(ftop));
    }
  }
  if(gtop != 0){
    for (auto x : g) {
      g0.push_back(x.OffSet(gtop));
      g1.push_back(x.OnSet0(gtop));
    }
  }

  // making disjoint
  std::vector<ZBDD> z;
  if(ftop < gtop) {
    z = disjointUnionVec(f, g1, dep+1);
    for (int i = 0; i < z.size(); i++) {
      z[i] = z[i].Change(gtop);
    }
    z = exorVecZdd(z, disjointUnionVec(f, g0,dep+1));
  }
  if(gtop < ftop) {
    z = disjointUnionVec(g, f1, dep+1);
    for (int i = 0; i < z.size(); i++) {
      z[i] = z[i].Change(ftop);
    }
    z = exorVecZdd(z, disjointUnionVec(g, f0, dep+1));
  }

  if(ftop == gtop) {
    z = exorVecZdd(disjointUnionVec(f0, g1, dep+1), disjointUnionVec(f1, g0, dep+1));
    for (int i = 0; i < z.size(); i++) {
      z[i] = z[i].Change(ftop);
    }
    z = exorVecZdd(z, disjointUnionVec(f0, g0, dep+1));
  };

// #ifndef NON_CACHE
//   cacheDJv.insert(std::make_pair(key, z));
//#endif
  
  return z;
}

//~~~~~
// ctoi ver.
boost::unordered_map<std::pair<bddword, bddword>, CtoI, Hash_UM> cacheDJJc; // Cache for Disjoint join (ctoi
CtoI disjointJoin(CtoI f, CtoI g){
  
  // check input and id
  int ftop = f.TopItem(), gtop = g.TopItem();
  bddword fid = f.GetZBDD().GetID(), gid = g.GetZBDD().GetID();
  
  if((ftop == 0 && f != CtoI()) || (gtop == 0 && g != CtoI())) {
    CtoI re;
    re = f*g; 
    return re;
  }
  if(ftop == 0 && f == CtoI()) {
    CtoI re = CtoI();  
    return re;
  }
  if(gtop == 0 && g == CtoI()) {
    CtoI re = CtoI(); 
    return re;
  }
  //=========
  if(fid > gid){
    CtoI temp;
    temp = f;
    f = g;
    g = temp;

    fid = f.GetZBDD().GetID();
    gid = g.GetZBDD().GetID();
    ftop = f.TopItem();
    gtop = g.TopItem();
  }

#ifndef NON_CACHE
  std::pair<bddword, bddword> key = std::pair<bddword, bddword> (fid, gid);
  //std::pair<bddword, bddword> key1 = std::pair<bddword, bddword> (fid, gid);
  //std::pair<bddword, bddword> key2 = std::pair<bddword, bddword> (gid, fid);
  if(cacheDJJc.find(key) != cacheDJJc.end()) {
    //std::cout << "key" << std::endl;
    return cacheDJJc[key];
  }
#endif
  //=========
  
  CtoI f0, f1, g0, g1;
  if(ftop != 0){
    f0 = f.Factor0(ftop);
    f1 = f.Factor1(ftop);
  }
  if(gtop != 0){
    g0 = g.Factor0(gtop);
    g1 = g.Factor1(gtop);
  }

  // making disjoint
  CtoI z;
  if(ftop < gtop) {
    z = disjointJoin(f, g1).AffixVar(gtop) + disjointJoin(f, g0);
  }
  if(gtop < ftop) {
    z = disjointJoin(g, f1).AffixVar(ftop) + disjointJoin(g, f0);
  }
  if(ftop == gtop) {
    z = (disjointJoin(f0, g1)+disjointJoin(f1, g0)).AffixVar(ftop) + disjointJoin(f0, g0);
  };

 #ifndef NON_CACHE
   cacheDJJc.insert(std::make_pair(key, z));
 #endif
  
  return z;
}
// focus set having few items
CtoI disjointJoin_MinItem(CtoI f, CtoI g){
  
  // check input and id
  bddword fid = f.GetZBDD().GetID(), gid = g.GetZBDD().GetID();
  int ftop = f.TopItem(), gtop = g.TopItem();
  if((ftop == 0 && f != CtoI()) || (gtop == 0 && g != CtoI())) {
    CtoI re;
    re = f*g; 
    return re;
  }
  if(ftop == 0 && f == CtoI()) {
    CtoI re = CtoI();  
    return re;
  }
  if(gtop == 0 && g == CtoI()) {
    CtoI re = CtoI(); 
    return re;
  }
  //=========
  CtoI fitem = f.Support().TotalVal(), gitem = g.Support().TotalVal();
  if(fitem.GetInt() > gitem.GetInt()){
    CtoI temp;
    temp = f;
    f = g;
    g = temp;

    fid = f.GetZBDD().GetID();
    gid = g.GetZBDD().GetID();
  }
  int mintop = f.TopItem();

#ifndef NON_CACHE
  std::pair<bddword, bddword> key = std::pair<bddword, bddword> (fid, gid);
  //std::pair<bddword, bddword> key1 = std::pair<bddword, bddword> (fid, gid);
  //std::pair<bddword, bddword> key2 = std::pair<bddword, bddword> (gid, fid);
  if(cacheDJJc.find(key) != cacheDJJc.end()) {
    //std::cout << "key" << std::endl;
    return cacheDJJc[key];
  }
#endif
  //=========
  
  CtoI f0, f1, g0, g1;
  if(mintop != 0){
    f0 = f.Factor0(mintop);
    f1 = f.Factor1(mintop);
  }
  if(mintop != 0){
    g0 = g.Factor0(mintop);
    g1 = g.Factor1(mintop);
  }

  // making disjoint
  CtoI z;
  z = (disjointJoin_MinItem(f0, g1)+disjointJoin_MinItem(f1, g0)).AffixVar(mintop) + disjointJoin_MinItem(f0, g0);
  

 #ifndef NON_CACHE
   cacheDJJc.insert(std::make_pair(key, z));
 #endif
  
  return z;
}

//~~~~~



// f @ g = a ((f_1 Join g_1) ∪ (f_0 @ g_1) ∪ (f_1 @ g_0)) ∪ (f_0 @ g_0)
// T @ T = F
// T @ F = F
// F @ T = F
// F @ F = F
ZBDD intersectionUnion(ZBDD f, ZBDD g){
  if(f.Top() == 0 || g.Top() == 0) return ZBDD(0);

  if(f.GetID() > g.GetID()){
    ZBDD temp;
    temp = f;
    f = g;
    g = temp;
  }
#ifndef NON_CACHE
  std::pair<bddword, bddword> key = std::pair<bddword, bddword> (f.GetID(), g.GetID());
  if(cacheJJ.find(key) != cacheJJ.end()) return cacheJJ[key];
#endif

  int ftop = f.Top();
  int gtop = g.Top();
  ZBDD f0 = f.OffSet(ftop);
  ZBDD f1 = f.OnSet0(ftop);
  ZBDD g0 = g.OffSet(gtop);
  ZBDD g1 = g.OnSet0(gtop);

  ZBDD z;
  if(ftop < gtop) z = intersectionUnion(f, g1).Change(gtop) + intersectionUnion(f, g0);
  if(gtop < ftop) z = intersectionUnion(g, f1).Change(ftop) + intersectionUnion(g, f0);
  if(gtop == ftop) z = (f1*g1 + intersectionUnion(f0, g1) + intersectionUnion(f1, g0)).Change(ftop) + intersectionUnion(f0, g0);

#ifndef NON_CACHE
  cacheJJ.insert(std::make_pair(key, z));
#endif
  return z;
}
