#pragma once

#include <tdzdd/DdSpec.hpp>
#include <tdzdd/DdSpecOp.hpp>
#include <tdzdd/DdStructure.hpp>
#include <tdzdd/DdEval.hpp>

#include "../../graph.cpp"
#include "convertEdgePermutation.cpp"

class RelaxedDegreeZDD: public tdzdd::PodArrayDdSpec<RelaxedDegreeZDD, unsigned char, 2> {
    int n;
    int m;
    int s;
    int t;
    Graph const G;
    std::vector<int> const perm;
    std::vector<int> permInv;
    std::vector<std::pair<int, int>> edgePermutation;
    std::vector<int> beginInd;
    std::vector<int> endInd;
    std::vector<std::vector<int>> frontier;
    std::vector<int> stateInd;
    std::vector<std::vector<int>> enterVertex;
    std::vector<std::vector<int>> leaveVertex;
    int maxMateSize;

public:
    RelaxedDegreeZDD(Graph G, std::vector<int> perm): G(G), perm(perm) {
        n = G.numVertices();
        m = G.numEdges();
        s = G.getStart();
        t = G.getTerminal();
        permInv.resize(n);
        for(int i = 0; i < n; i++) permInv[perm[i]] = i;

        edgePermutation = convertEdgePermutation(G, perm);
        beginInd.assign(n, -1);
        endInd.assign(n, 0);
        for(int i = 0; i < m; i++) {
            int u = edgePermutation[i].first;
            int v = edgePermutation[i].second;
            if(beginInd[u] == -1) beginInd[u] = i;
            if(beginInd[v] == -1) beginInd[v] = i;
            endInd[u] = i;
            endInd[v] = i;
        }

        frontier.resize(m);
        for(int i = 0; i < n; i++) {
            for(int j = beginInd[i]; j <= endInd[i]; j++) {
                frontier[j].push_back(i);
            }
        }

        maxMateSize = 0;
        for(int i = 0; i < m; i++) {
            if(maxMateSize < frontier[i].size()) {
                maxMateSize = frontier[i].size();
            }
        }

        enterVertex.resize(m);
        leaveVertex.resize(m);
        for(int i = 0; i < n; i++) {
            enterVertex[beginInd[i]].push_back(i);
            leaveVertex[endInd[i]].push_back(i);
        }

        stateInd.assign(n, -1);
        std::vector<int> stateTmp(n, -1);
        for(int i = 0; i < m; i++) {
            for(auto t: enterVertex[i]) {
                for(int j = 0; j < n; j++) {
                    if(stateTmp[j] == -1) {
                        stateTmp[j] = t;
                        stateInd[t] = j;
                        break;
                    }
                }
            }
            if(i == 0) continue;
            for(auto t: leaveVertex[i - 1]) {
                for(int j = 0; j < n; j++) {
                    if(stateTmp[j] == t) {
                        stateTmp[j] = -1;
                        break;
                    }
                }
            }
        }
        setArraySize(maxMateSize + 1);
    }

    int getComp(unsigned char* state, int level) const {
        return state[stateInd[level]];
    }

    int setComp(unsigned char* state, int level, int value) const {
        return state[stateInd[level]] = value;
    }

    int getRoot(unsigned char* mate) const {
        for(int i = 0; i <= maxMateSize; i++) {
            mate[i] = 0;
        }
        return m;
    }

    int getChild(unsigned char* mate, int level, int value) const {
        int u = edgePermutation[m - level].first;
        int v = edgePermutation[m - level].second;
        for(auto e: enterVertex[m - level]) {
            setComp(mate, e, 0);
        }
        if(value == 1) {
            int mate_u = getComp(mate, u);
            int mate_v = getComp(mate, v);
            setComp(mate, u, !mate_u);
            setComp(mate, v, !mate_v);
        }
        for(auto e: leaveVertex[m - level]) {
            int state = getComp(mate, e);
            if(e == s || e == t) {
                if(state != 1) return 0; 
            }
            else if(state == 1) return 0;
            setComp(mate, e, 0);
        }
        level--;
        if(level == 0) return -1;
        return level;
    }
};
